package labss.lang.il.visitors;

import java.io.PrintStream;

import labss.lang.il.model.Expression;
import labss.lang.il.model.ILEntity;
import labss.lang.il.model.impl.Addition;
import labss.lang.il.model.impl.And;
import labss.lang.il.model.impl.Assignment;
import labss.lang.il.model.impl.BinaryExpression;
import labss.lang.il.model.impl.Block;
import labss.lang.il.model.impl.Bool;
import labss.lang.il.model.impl.Division;
import labss.lang.il.model.impl.Empty;
import labss.lang.il.model.impl.Equals;
import labss.lang.il.model.impl.GreaterThan;
import labss.lang.il.model.impl.If;
import labss.lang.il.model.impl.IfElse;
import labss.lang.il.model.impl.Int;
import labss.lang.il.model.impl.LessThan;
import labss.lang.il.model.impl.Multiplication;
import labss.lang.il.model.impl.Not;
import labss.lang.il.model.impl.NullExpression;
import labss.lang.il.model.impl.Or;
import labss.lang.il.model.impl.Subtraction;
import labss.lang.il.model.impl.Variable;
import labss.lang.il.model.impl.While;

/**
 * @author Riccardo Solmi
 */
public class IndentedPrettyPrintVisitor implements ILVisitor {
	protected final PrintStream out;
	private final String indentationString = "                                ";
	protected int indentation = 0;
	protected int indentationSize = 4;

	public IndentedPrettyPrintVisitor(PrintStream out) {
		this.out = out;
	}

	public static void prettyPrint(ILEntity entity, PrintStream out) {
		entity.accept(new IndentedPrettyPrintVisitor(out));
	}

	protected String indentationString() {
		if (indentation > 0) {
			if (indentation * indentationSize < indentationString.length())
				return indentationString.substring(0, indentation * indentationSize);
			else
				return indentationString;
		} else
			return "";
	}
	protected void printIndentation() {
		if (indentation > 0) {
			if (indentation * indentationSize < indentationString.length())
				out.print(indentationString.substring(0, indentation * indentationSize));
			else
				out.print(indentationString);
		}
	}

	protected void visitChild(Expression parent, Expression child) {
/*		if (parent.hasPrecedence(child)) {
			out.print("(");
			child.accept(this);
			out.print(")");
		} else
*/			child.accept(this);
	}
	protected void visitBinExp(BinaryExpression exp, String opStr) {
		visitChild(exp, exp.getExp1());
		out.print(opStr);
		visitChild(exp, exp.getExp2());
	}

	public void visit(Empty exp) {
		printIndentation();
		out.println(";");
	}
	public void visit(NullExpression exp) {
		out.print("null");
	}

	public void visit(While stm) {
		printIndentation();
		out.print("while (");
		stm.getCondition().accept(this);
		indentation++;
		if (stm.getTrueBody() instanceof Block) {
			out.println(") {");
			stm.getTrueBody().accept(this);
			indentation--;
			printIndentation();
			out.println("}");
		} else {
			out.println(")");
			stm.getTrueBody().accept(this);
			indentation--;
		}
	}

	public void visit(If stm) {
		printIndentation();
		out.print("if (");
		stm.getCondition().accept(this);
		indentation++;
		if (stm.getTrueBody() instanceof Block) {
			out.println(") {");
			stm.getTrueBody().accept(this);
			out.println("}");
		} else {
			out.println(")");
			stm.getTrueBody().accept(this);
		}
		indentation--;
	}
	public void visit(IfElse stm) {
		printIndentation();
		out.print("if (");
		stm.getCondition().accept(this);
		indentation++;
		if (stm.getTrueBody() instanceof Block) {
			out.println(") {");
			stm.getTrueBody().accept(this);
			out.println("}");
		} else {
			out.println(")");
			stm.getTrueBody().accept(this);
		}
		indentation--;

		printIndentation();
		indentation++;
		if (stm.getFalseBody() instanceof Block) {
			out.println("else {");
			stm.getFalseBody().accept(this);
			out.println("}");
		} else {
			out.println("else");
			stm.getFalseBody().accept(this);
		}
		indentation--;
	}

	public void visit(Assignment stm) {
		printIndentation();
		stm.getVariable().accept(this);
		out.print(" = ");
		stm.getExp().accept(this);
		out.println(";");
	}

	public void visit(Block stm) {
		for (int i=0; i<stm.size(); i++)
			stm.get(i).accept(this);
	}

	public void visit(Bool exp) {
		out.print(exp);
	}
	public void visit(Int exp) {
		out.print(exp);
	}

	public void visit(Variable exp) {
		out.print(exp.getName());
	}

	public void visit(Addition exp) {
		visitBinExp(exp, " + ");
	}
	public void visit(Subtraction exp) {
		visitBinExp(exp, " - ");
	}
	public void visit(Multiplication exp) {
		visitBinExp(exp, " * ");
	}
	public void visit(Division exp) {
		visitBinExp(exp, " / ");
	}

	public void visit(Equals exp) {
		visitBinExp(exp, " == ");
	}
	public void visit(GreaterThan exp) {
		visitBinExp(exp, " > ");
	}
	public void visit(LessThan exp) {
		visitBinExp(exp, " < ");
	}

	public void visit(And exp) {
		visitBinExp(exp, " && ");
	}
	public void visit(Or exp) {
		visitBinExp(exp, " || ");
	}
	public void visit(Not exp) {
		out.print("!");
		visitChild(exp, exp.getExp());
	}
}
