package labss.lang.il.visitors;

import java.util.HashMap;
import java.util.Map;

import labss.lang.il.factories.ILAbstractFactory;
import labss.lang.il.model.Expression;
import labss.lang.il.model.ILEntity;
import labss.lang.il.model.Literal;
import labss.lang.il.model.Statement;
import labss.lang.il.model.impl.Addition;
import labss.lang.il.model.impl.And;
import labss.lang.il.model.impl.Assignment;
import labss.lang.il.model.impl.BinaryExpression;
import labss.lang.il.model.impl.Block;
import labss.lang.il.model.impl.Bool;
import labss.lang.il.model.impl.Division;
import labss.lang.il.model.impl.Empty;
import labss.lang.il.model.impl.Equals;
import labss.lang.il.model.impl.GreaterThan;
import labss.lang.il.model.impl.If;
import labss.lang.il.model.impl.IfElse;
import labss.lang.il.model.impl.Int;
import labss.lang.il.model.impl.LessThan;
import labss.lang.il.model.impl.Multiplication;
import labss.lang.il.model.impl.Not;
import labss.lang.il.model.impl.NullExpression;
import labss.lang.il.model.impl.Or;
import labss.lang.il.model.impl.Subtraction;
import labss.lang.il.model.impl.Variable;
import labss.lang.il.model.impl.While;


/**
 * @author Riccardo Solmi
 */
public class InterpreterVisitor implements ILVisitor {
	private BindingManager bindingManager = BindingManager.create();
	private Literal result = null;
	
	protected InterpreterVisitor() {
	}

	public static BindingManager interpret(ILEntity program) {
		InterpreterVisitor v = new InterpreterVisitor();
		program.accept(v);
		return v.bindingManager;
	}

	private final Literal evaluate(Expression exp) {
		exp.accept(this);
		return result;
	}
	private final void execute(Statement stm) {
		stm.accept(this);
	}

	public void visit(Empty stm) {
	}

	public void visit(NullExpression stm) {
		throw new IllegalStateException("Illegal method call on a null object");
	}

	public void visit(While stm) {
		while (evaluate(stm.getCondition()).booleanValue())
			execute(stm.getTrueBody());
	}

	public void visit(If stm) {
		if (evaluate(stm.getCondition()).booleanValue())
			execute(stm.getTrueBody());
	}
	public void visit(IfElse stm) {
		if (evaluate(stm.getCondition()).booleanValue())
			execute(stm.getTrueBody());
		else
			execute(stm.getFalseBody());
	}

	public void visit(Assignment stm) {
		bindingManager.bind(stm.getVariable().getName(), evaluate(stm.getExp()));
	}

	public void visit(Block stm) {
		for (int i=0; i<stm.size(); i++)
			stm.get(i).accept(this);
	}

	public void visit(Bool exp) {
		result = exp;
	}
	public void visit(Int exp) {
		result = exp;
	}

	public void visit(Variable exp) {
		result = bindingManager.getValue(exp.getName());
		if (result == null)
			throw new IllegalStateException("Variabile non inizializzata: "+exp.getName());
	}

	private final Literal createLiteral(boolean value) {
		return ILAbstractFactory.instance.createBoolean(value);
	}
	private final Literal createLiteral(int value) {
		return ILAbstractFactory.instance.createInt(value);
	}

	private final void setResult(int value) {
		result = createLiteral(value);
	}
	private final int intValue(Expression exp) {
		return evaluate(exp).intValue();
	}

	static interface IxItoI {
		public int eval(int a, int b);
	}
	private final void evaluate(BinaryExpression exp, IxItoI op) {
		result = createLiteral(op.eval(evaluate(exp.getExp1()).intValue(), evaluate(exp.getExp2()).intValue()));
	}

	public void visit(Addition exp) {
		setResult(intValue(exp.getExp1()) + intValue(exp.getExp2()));
	}

	public void visit(Subtraction exp) {
		evaluate(exp, new IxItoI() {
			public int eval(int a, int b) {
				return a - b;
			}
		});
	}

	public void visit(Multiplication exp) {
		result = createLiteral(evaluate(exp.getExp1()).intValue() * evaluate(exp.getExp2()).intValue());
	}

	public void visit(Division exp) {
		result = createLiteral(evaluate(exp.getExp1()).intValue() / evaluate(exp.getExp2()).intValue());
	}

	public void visit(Equals exp) {
		result = createLiteral(evaluate(exp.getExp1()).intValue() == evaluate(exp.getExp2()).intValue());
	}
	public void visit(GreaterThan exp) {
		result = createLiteral(evaluate(exp.getExp1()).intValue() > evaluate(exp.getExp2()).intValue());
	}
	public void visit(LessThan exp) {
		result = createLiteral(evaluate(exp.getExp1()).intValue() < evaluate(exp.getExp2()).intValue());
	}

	public void visit(And exp) {
		result = createLiteral(evaluate(exp.getExp1()).booleanValue() && evaluate(exp.getExp2()).booleanValue());
	}
	public void visit(Or exp) {
		result = createLiteral(evaluate(exp.getExp1()).booleanValue() || evaluate(exp.getExp2()).booleanValue());
	}

	public void visit(Not exp) {
		result = createLiteral(!evaluate(exp.getExp()).booleanValue());
	}

	
	public static class BindingManager {
		public static BindingManager create() {
			return new BindingManager() {
				public void exitScope() {
					throw new IllegalStateException("exitScope from top level");		
				}

				public boolean contains(String identifier) {
					return map.containsKey(identifier);
				}
				public Literal getValue(String identifier) {
					return map.get(identifier);
				}
			};
		}

		private BindingManager enclosingScope;
		protected Map<String, Literal> map = new HashMap<String, Literal>();

		private BindingManager() {
		}
		private BindingManager(BindingManager enclosingScope, Map<String, Literal> map) {
			this.enclosingScope = enclosingScope;
			this.map = map;
		}
		
		public void enterScope() {
			this.enclosingScope = new BindingManager(enclosingScope, map);
			this.map = new HashMap<String, Literal>();
		}
		public void exitScope() {
			this.enclosingScope = enclosingScope.enclosingScope;
			this.map = enclosingScope.map;
		}

		public boolean contains(String identifier) {
			return map.containsKey(identifier) ? true : enclosingScope.contains(identifier);
		}
		public void bind(String identifier, Literal value) {
			map.put(identifier, value);
		}
		public Literal getValue(String identifier) {
			Literal value = map.get(identifier);
			return (value != null) ? value : enclosingScope.getValue(identifier);
		}
	}
}
